import asyncio
import queue
from queue import Queue
from typing import TypeVar

import sanic.response as sanic_response
from pydantic import BaseModel

T = TypeVar("T")

timeout = 10


class ErrorCode(BaseModel):
    status: int = 200
    code: int
    message: str


class ErrorCodes:
    ok = ErrorCode(code=20000, message="OK")
    invalid_json = ErrorCode(code=40001, message="请求JSON格式错误")
    username_existed = ErrorCode(code=40002, message="用户名已存在")
    invalid_username_or_password = ErrorCode(code=40003, message="用户名或密码不正确")
    unauthorized = ErrorCode(code=40100, message="未登录", status=401)
    invalid_token = ErrorCode(code=40101, message="无效的令牌", status=401)
    file_size_exceeded = ErrorCode(code=40102, message="文件尺寸超出限制，请将文件尺寸控制在3MB以内")
    empty_file = ErrorCode(code=40103, message="上传的文件为空，请检查后重新上传")
    file_count_exceeded = ErrorCode(code=40104, message="上传文件数量超过限制，每个知识库最多上传5个文件")
    file_extension_invalid = ErrorCode(code=40105, message="暂不支持该文件类型")
    permission_denied = ErrorCode(code=40300, message="你没有权限执行此操作")
    internal_server_error = ErrorCode(code=50000, message="服务器内部错误", status=500)
    knowledge_base_not_found = ErrorCode(code=40400, message="知识库不存在")


def response(code: ErrorCode = ErrorCodes.ok, data: T = ""):
    if isinstance(data, BaseModel):
        data = data.dict()
    return sanic_response.json({"code": code.code, "data": data, "message": code.message}, status=code.status)


def stream(q: Queue):
    async def stream_from_queue(res: sanic_response.StreamingHTTPResponse):
        total_sleep = 0
        while True:
            if total_sleep >= timeout:
                raise TimeoutError()
            try:
                data = q.get_nowait()
                if data == "end":
                    break
                await res.write(data)
                print(f"writen the data {data}")
            except queue.Empty:
                await asyncio.sleep(0.5)
                total_sleep += 0.5
        print("receive end")

    return sanic_response.stream(stream_from_queue)



import re
from typing import List, Optional, Dict

from pydantic import Field

from lemon_rag.dependencies.vector_access import vector_access, DocSearchRes
from lemon_rag.lemon_runtime import models


class RelatedSentence(DocSearchRes):
    file: models.KnowledgeBaseFileTab
    content: List[str] = Field(default_factory=list)


def retrieve_related_paragraphs(query: str, knowledgebase_id_list: List[int]):
    keys = re.split(r",.，。\n", query)
    search_res = vector_access.find_sentence(keys, knowledgebase_id_list)

    id_to_ref_mapping: Dict[int, RelatedSentence] = {}
    for hit in search_res:
        sentence: models.KnowledgeSentence = models.KnowledgeSentence.get_or_none(hit.id)
        if sentence is None:
            continue
        paragraph = sentence.paragraph
        current_ref_content = paragraph.raw_content
        if sentence.index <= paragraph.total_sentences // 3:
            former_paragraph: Optional[models.KnowledgeParagraph] = models.KnowledgeParagraph.get_or_none(
                file=paragraph.file, index=paragraph.index - 1)
            if former_paragraph is not None:
                current_ref_content = former_paragraph.raw_content + current_ref_content
        elif sentence.index >= paragraph.total_sentences * 2 // 3:
            # 添加后一段
            next_paragraph: Optional[models.KnowledgeParagraph] = models.KnowledgeParagraph.get_or_none(
                file=paragraph.file, index=paragraph.index + 1)
            if next_paragraph is not None:
                current_ref_content += next_paragraph.raw_content

        rp = id_to_ref_mapping.get(
            paragraph.file.id,
            RelatedSentence(file=paragraph.file, doc_id=hit.doc_id, id=hit.id)
        )
        rp.content.append(current_ref_content)
        id_to_ref_mapping[paragraph.file.id] = rp

    return id_to_ref_mapping



import traceback

from lemon_rag.api.local import get_rid, get_user


def lemon_info(value: str):
    pass

def info(msg: str, *args):
    suffix = f" rid={get_rid()} user_id={get_user()}"
    try:
        log_expr = (msg + suffix) % args
        print(log_expr)
        lemon_info(log_expr)
    except Exception as e:
        print(traceback.format_exc())
        lemon_info(traceback.format_exc())



import sanic
from sanic.compat import Header
from sanic.helpers import STATUS_CODES

import sanic.response as sanic_response
import sanic.asgi

class StreamingHTTPResponse(sanic_response.HTTPResponse):

    def __init__(
            self,
            streaming_fn,
            status=200,
            headers=None,
            content_type="text/plain",
            chunked=True,
    ):
        self.content_type = content_type
        self.streaming_fn = streaming_fn
        self.status = status
        self.headers = Header(headers or {})
        self.chunked = chunked
        self._cookies = None
        super().__init__()

    async def write(self, data):
        """Writes a chunk of data to the streaming response.

        :param data: bytes-ish data to be written.
        """
        if type(data) != bytes:
            data = self._encode_body(data)

        if self.chunked:
            await self.protocol.push_data(b"%x\r\n%b\r\n" % (len(data), data))
        else:
            await self.protocol.push_data(data)
        await self.protocol.drain()

    async def stream(
            self, version="1.1", keep_alive=False, keep_alive_timeout=None
    ):
        """Streams headers, runs the `streaming_fn` callback that writes
        content to the response body, then finalizes the response body.
        """
        if version != "1.1":
            self.chunked = False
        headers = self.get_headers(
            version,
            keep_alive=keep_alive,
            keep_alive_timeout=keep_alive_timeout,
        )
        await self.protocol.push_data(headers)
        await self.protocol.drain()
        await self.streaming_fn(self)
        if self.chunked:
            await self.protocol.push_data(b"0\r\n\r\n")
        # no need to await drain here after this write, because it is the
        # very last thing we write and nothing needs to wait for it.

    def get_headers(
            self, version="1.1", keep_alive=False, keep_alive_timeout=None
    ):
        # This is all returned in a kind-of funky way
        # We tried to make this as fast as possible in pure python
        timeout_header = b""
        if keep_alive and keep_alive_timeout is not None:
            timeout_header = b"Keep-Alive: %d\r\n" % keep_alive_timeout

        if self.chunked and version == "1.1":
            self.headers["Transfer-Encoding"] = "chunked"
            self.headers.pop("Content-Length", None)
        self.headers["Content-Type"] = self.headers.get(
            "Content-Type", self.content_type
        )

        headers = self._parse_headers()

        if self.status == 200:
            status = b"OK"
        else:
            status = STATUS_CODES.get(self.status)

        return (b"HTTP/%b %d %b\r\n" b"%b" b"%b\r\n") % (
            version.encode(),
            self.status,
            status,
            timeout_header,
            headers,
        )


setattr(sanic_response, "StreamingHTTPResponse", StreamingHTTPResponse)
setattr(sanic.asgi, "StreamingHTTPResponse", StreamingHTTPResponse)


import requests
# url = "https://lemon.lemonstudio.tech:8443/50999ced2d425236b3fdb814c674e6b4test/f782d741bf2d505f825662e1361a846f/restful/v1/hello_stream"
# url = "http://localhost:9999"
url = "http://localhost:7001/restful/v1/hello_stream"
with requests.post(url, headers={"Accept-Encoding": "identity", "X-Accel-Buffering": "no"}, stream = True) as resp:
    for line in resp.iter_content():
        if line:
            print(line)
    print(resp.headers)

"curl -X POST -N 'http://localhost:7001/restful/v1/hello_stream'"
"curl -X POST -N 'http://localhost:7000/50999ced2d425236b3fdb814c674e6b4test/f782d741bf2d505f825662e1361a846f/restful/v1/hello_stream'"
"curl -X POST -N 'http://localhost:7000/50999ced2d425236b3fdb814c674e6b4test/f782d741bf2d505f825662e1361a846f/restful/v1/hello_stream'"



import abc
import io
import re
from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel


class FileExtension(str, Enum):
    doc = "doc"
    docx = "docx"
    pdf = "pdf"
    txt = "txt"

    @classmethod
    def _missing_(cls, value: str) -> 'FileExtension':
        value = value.lower()
        return FileExtension(value)


class Paragraph(BaseModel):
    index: int = 0
    sentences: List[str] = []


class ContentExtractor(abc.ABC):
    @abc.abstractmethod
    def extract_content(self, file_path: Union[str, io.IOBase]) -> List[Paragraph]:
        pass


def split_text(document: str) -> List[str]:
    delimiter_pattern = r'([\n。！？.!?])'
    segments = re.split(delimiter_pattern, document)
    sentences = [segments[i] + segments[i + 1] for i in range(0, len(segments) - 1, 2)]
    return sentences


class TxtExtractor(ContentExtractor):
    paragraph_threshold = 300  # paragraph 必须是有换行分割、符号分割
    sentence_threshold = 30

    def extract_content(self, file_path: Union[str, io.IOBase]) -> List[Paragraph]:
        if isinstance(file_path, io.IOBase):
            content = file_path.read()
        else:
            with open(file_path, "r") as f:
                content = f.read()
        res: List[Paragraph] = []
        current_paragraph: Optional[Paragraph] = None

        for sentence in split_text(content):
            if not current_paragraph or len("".join(current_paragraph.sentences)) >= self.paragraph_threshold:
                current_paragraph = Paragraph()
                res.append(current_paragraph)
            current_paragraph.sentences.append(sentence)
        return res



from enum import Enum
from typing import Optional, List, Any

from pydantic import BaseModel, Field


class KnowledgeBase(BaseModel):
    id: int
    name: str


class ChatRole(str, Enum):
    assistant = "assistant"
    notification_center = "notification_center"
    user = "user"


class Session(BaseModel):
    id: int
    topic: str
    title: str
    messages: Optional[List['Message']]
    latest_msg_ts: Optional[int]
    assistant_role: str
    create_at: int
    version: int


class Message:
    msg_id: int
    content: str
    text: str


class CardComponentType(str, Enum):
    text = "text"
    project_card = "project_card"
    reference = "reference"
    audio = "audio"
    expense_trace_card = "expense_trace_card"
    file = "file"
    guide_card = "guide_card"


class CardComponent(BaseModel):
    type: CardComponentType
    data: Any


class CardMessage(Message):
    title: str = ""
    components: List[CardComponent] = Field(default_factory=list)


class MessageChunkAction(int, Enum):
    append_text = 1


class ResponseChunk(BaseModel):
    action: MessageChunkAction


class AppendText(ResponseChunk):
    action: MessageChunkAction = MessageChunkAction.append_text
    text: str


class AddRef(ResponseChunk):
    sentence_id: int
    paragraph_id: int
    file_id: int
    filename: str



import queue
from concurrent.futures import ThreadPoolExecutor
from typing import Generator, Iterable

streaming_pool = ThreadPoolExecutor(max_workers=16)


def submit_streaming_task(generator: Iterable[str]) -> queue.Queue:
    q = queue.Queue(maxsize=128)

    def inner_task():
        for value in generator:
            print("iterate the generator", value)
            q.put(value)
        q.put("end")

    streaming_pool.submit(inner_task)
    return q
